@@ -24,6 +24,7 @@ import (
 	"crypto/tls"
 	"fmt"
 	"github.com/IrineSistiana/mosdns/v4/pkg/dnsutils"
+	"github.com/IrineSistiana/mosdns/v4/pkg/upstream/bootstrap"
 	"github.com/IrineSistiana/mosdns/v4/pkg/upstream/doh"
 	"github.com/IrineSistiana/mosdns/v4/pkg/upstream/h3roundtripper"
 	"github.com/IrineSistiana/mosdns/v4/pkg/upstream/transport"
@@ -67,12 +68,11 @@ type Opt struct {
 	// Not implemented for udp upstreams and doh upstreams with http/3.
 	Socks5 string
 
-	// SoMark specifies the mark for each packet sent through this upstream.
+	// SoMark sets the socket SO_MARK option in unix system.
 	SoMark int
 
-	// Interface specifies the system network interface for this upstream to send
-	// data.
-	Interface string
+	// BindToDevice sets the socket SO_BINDTODEVICE option in unix system.
+	BindToDevice string
 
 	// IdleTimeout specifies the idle timeout for long-connections.
 	// Available for TCP, DoT, DoH.
@@ -90,9 +90,17 @@ type Opt struct {
 	// MaxConns limits the total number of connections, including connections
 	// in the dialing states.
 	// Implemented for TCP/DoT pipeline enabled upstreams and DoH upstreams.
-	// Default is 1.
+	// Default is 2.
 	MaxConns int
 
+	// Bootstrap specifies a plain dns server for the go runtime to solve the
+	// domain of the upstream server. It SHOULD be an IP address. Custom port
+	// is supported.
+	// Note: Use a domain address may cause dead resolve loop and additional
+	// latency to dial upstream server.
+	// HTTP3 is not supported.
+	Bootstrap string
+
 	// TLSConfig specifies the tls.Config that the TLS client will use.
 	// Available for DoT, DoH upstreams.
 	TLSConfig *tls.Config
@@ -115,16 +123,22 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 		return nil, fmt.Errorf("invalid server address, %w", err)
 	}
 
+	dialer := &net.Dialer{
+		Resolver: bootstrap.NewPlainBootstrap(opt.Bootstrap),
+		Control: getSocketControlFunc(socketOpts{
+			so_mark:        opt.SoMark,
+			bind_to_device: opt.BindToDevice,
+		}),
+	}
+
 	switch addrURL.Scheme {
 	case "", "udp":
 		dialAddr := getDialAddrWithPort(addrURL.Host, opt.DialAddr, 53)
-		ut := &transport.Transport{
+
+		uto := transport.Opts{
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
-				d := net.Dialer{
-					Control: getSetMarkFunc(opt.SoMark),
-				}
-				return d.DialContext(ctx, "udp", dialAddr)
+				return dialer.DialContext(ctx, "udp", dialAddr)
 			},
 			WriteFunc: dnsutils.WriteMsgToUDP,
 			ReadFunc: func(c io.Reader) (*dns.Msg, int, error) {
@@ -134,27 +148,32 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 			MaxConns:       opt.MaxConns,
 			IdleTimeout:    time.Second * 60,
 		}
-		tt := &transport.Transport{
+		ut, err := transport.NewTransport(uto)
+		if err != nil {
+			return nil, fmt.Errorf("cannot init udp transport, %w", err)
+		}
+		tto := transport.Opts{
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
-				d := net.Dialer{
-					Control: getSetMarkFunc(opt.SoMark),
-				}
-				return d.DialContext(ctx, "tcp", dialAddr)
+				return dialer.DialContext(ctx, "tcp", dialAddr)
 			},
 			WriteFunc: dnsutils.WriteMsgToTCP,
 			ReadFunc:  dnsutils.ReadMsgFromTCP,
 		}
+		tt, err := transport.NewTransport(tto)
+		if err != nil {
+			return nil, fmt.Errorf("cannot init tcp transport, %w", err)
+		}
 		return &udpWithFallback{
 			u: ut,
 			t: tt,
 		}, nil
 	case "tcp":
 		dialAddr := getDialAddrWithPort(addrURL.Host, opt.DialAddr, 53)
-		t := &transport.Transport{
+		to := transport.Opts{
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
-				return dialTCP(ctx, dialAddr, opt.Socks5, opt.SoMark)
+				return dialTCP(ctx, dialAddr, opt.Socks5, dialer)
 			},
 			WriteFunc:      dnsutils.WriteMsgToTCP,
 			ReadFunc:       dnsutils.ReadMsgFromTCP,
@@ -162,7 +181,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 			EnablePipeline: opt.EnablePipeline,
 			MaxConns:       opt.MaxConns,
 		}
-		return t, nil
+		return transport.NewTransport(to)
 	case "tls":
 		var tlsConfig *tls.Config
 		if opt.TLSConfig != nil {
@@ -175,10 +194,10 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 		}
 
 		dialAddr := getDialAddrWithPort(addrURL.Host, opt.DialAddr, 853)
-		t := &transport.Transport{
+		to := transport.Opts{
 			Logger: opt.Logger,
 			DialFunc: func(ctx context.Context) (net.Conn, error) {
-				conn, err := dialTCP(ctx, dialAddr, opt.Socks5, opt.SoMark)
+				conn, err := dialTCP(ctx, dialAddr, opt.Socks5, dialer)
 				if err != nil {
 					return nil, err
 				}
@@ -195,7 +214,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 			EnablePipeline: opt.EnablePipeline,
 			MaxConns:       opt.MaxConns,
 		}
-		return t, nil
+		return transport.NewTransport(to)
 	case "https":
 		idleConnTimeout := time.Second * 30
 		if opt.IdleTimeout > 0 {
@@ -210,7 +229,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 		var t http.RoundTripper
 		var addonCloser io.Closer // udpConn
 		if opt.EnableHTTP3 {
-			lc := net.ListenConfig{Control: getSetMarkFunc(opt.SoMark)}
+			lc := net.ListenConfig{Control: getSocketControlFunc(socketOpts{so_mark: opt.SoMark, bind_to_device: opt.BindToDevice})}
 			conn, err := lc.ListenPacket(context.Background(), "udp", "")
 			if err != nil {
 				return nil, fmt.Errorf("failed to init udp socket for quic")
@@ -227,7 +246,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 					MaxConnectionReceiveWindow:     64 * 1024,
 				},
 				DialFunc: func(ctx context.Context, _ string, tlsCfg *tls.Config, cfg *quic.Config) (quic.EarlyConnection, error) {
-					ua, err := net.ResolveUDPAddr("udp", dialAddr)
+					ua, err := net.ResolveUDPAddr("udp", dialAddr) // TODO: Support bootstrap.
 					if err != nil {
 						return nil, err
 					}
@@ -237,7 +256,7 @@ func NewUpstream(addr string, opt *Opt) (Upstream, error) {
 		} else {
 			t1 := &http.Transport{
 				DialContext: func(ctx context.Context, network, _ string) (net.Conn, error) { // overwrite server addr
-					return dialTCP(ctx, dialAddr, opt.Socks5, opt.SoMark)
+					return dialTCP(ctx, dialAddr, opt.Socks5, dialer)
 				},
 				TLSClientConfig:     opt.TLSConfig,
 				TLSHandshakeTimeout: tlsHandshakeTimeout,
